# -*- coding: utf-8 -*-
"""
Code that goes along with the 'user guide' in the BVR docs.

@author: Rebecca Kohlhaas
"""

import copy
import numpy as np
import pandas as pd
import sys
import joblib

import matplotlib as mpl
mpl.rcParams.update(mpl.rcParamsDefault)

# Add BayesValidRox path
sys.path.append("../../src/")
from bayesvalidrox import Input, ExpDesigns, PyLinkForwardModel, MetaModel, Engine, PostProcessing, Discrepancy, BayesInference, BayesModelComparison

if __name__ == '__main__':
    #### Priors, input space and experimental design
    version = 1
    Inputs = Input()
    
    # Version 1
    if version == 1:
        Inputs.add_marginals()
        Inputs.Marginals[0].name = '$X$'    
        Inputs.Marginals[0].dist_type = 'unif'
        Inputs.Marginals[0].parameters =  [-5, 5]
        
    # Version 2
    if version == 2:
        inputParams = np.random.uniform(-5,5,100)
        Inputs.add_marginals()
        Inputs.Marginals[0].name = '$X$'
        Inputs.Marginals[0].input_data = inputParams

    # Create ExpDesign and generate samples
    ExpDesign = ExpDesigns(Inputs)
    ExpDesign.sampling_method = 'latin_hypercube'
    samples = ExpDesign.generate_samples(10)
    ExpDesign.plot_samples(samples)
    
    #### Models
    x_values = np.arange(0,1,0.1)
    Model = PyLinkForwardModel()
    Model.link_type = 'Function'
    Model.py_file = 'model'
    Model.name = 'model'
    Model.Output.names = ['A']#, 'B']
    Model.func_args = {'x_values': x_values}
    Model.store = False
    
    output, samples = Model.run_model_parallel(samples, mp = True)
    
    #from model import model
    #out1 = model(samples)
    
    #### Training surrogate models
    MetaMod = MetaModel(Inputs)
    MetaMod.meta_model_type = 'aPCE'
    MetaMod.pce_reg_method = 'FastARD'
    MetaMod.pce_deg = 3
    MetaMod.pce_q_norm = 1
    
    # TODO: add check in Metamod/ExpDesign to ensure that n_init_samples matches the length of the given .X!
    ExpDesign.n_init_samples = 30
    ExpDesign.sampling_method = 'random'#'user'
    #ExpDesign.X = samples
    
    Engine_ = Engine(MetaMod, Model, ExpDesign)
    Engine_.start_engine()
    Engine_.train_normal()
    
    mean, stdev = Engine_.eval_metamodel(nsamples = 10)
    mean, stdev = Engine_.MetaModel.eval_metamodel(samples)
    
    #### Active learning
    if 0:
        ExpDesign.n_new_samples = 1
        ExpDesign.n_max_samples = 14
        ExpDesign.mod_LOO_threshold = 1e-16
        
        ExpDesign.tradeoff_scheme = None
        
        ExpDesign.explore_method = 'random'
        ExpDesign.n_canddidate = 1000
        ExpDesign.n_cand_groups = 4
        
        ExpDesign.exploit_method = 'VarOptDesign'
        ExpDesign.util_func = 'EIGF'
        
        Engine_ = Engine(copy.deepcopy(MetaMod), Model, copy.deepcopy(ExpDesign))
        Engine_.train_sequential()
    
    #### Postprocessing
    PostProc = PostProcessing(Engine_)
    PostProc.valid_metamodel(n_samples=1)
    PostProc.check_accuracy(n_samples=10)
    PostProc.plot_moments()
    PostProc.sobol_indices()
    #PostProc.plot_seq_design_diagnostics()
    
    # Sanity check - test on training data
    mean, stdev = Engine_.eval_metamodel(Engine_.ExpDesign.X)
    print(mean['A']-Engine_.ExpDesign.Y['A'])
    
    #### BayesInference
    
    true_sample = [[2]]
    observation = Model.run_model_parallel(true_sample)[0]
    Model.observations = {}
    for key in observation:
        if key == 'x_values':
            Model.observations[key]=observation[key]
        else:
            Model.observations[key]=observation[key][0]
            
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2 + 0.01
    
    BayesObj = BayesInference(Engine_)
    BayesObj.emulator = True
    BayesObj.Discrepancy = DiscrepancyOpts
    BayesObj.plot_post_pred = True
    
    BayesObj.inference_method = 'MCMC'
    import emcee
    BayesObj.mcmc_params = {
        'n_steps': 100, #e4,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
        }
    BayesObj.create_inference()
    
    
    #### Model Comparison
    Model1 = PyLinkForwardModel()
    Model1.link_type = 'Function'
    Model1.py_file = 'model'
    Model1.name = 'model'
    Model1.Output.names = ['B']
    Model1.func_args = {'x_values': x_values}
    Model1.store = False
    
    MetaMod1 = MetaModel(Inputs)
    MetaMod1.meta_model_type = 'aPCE'
    MetaMod1.pce_reg_method = 'FastARD'
    MetaMod1.pce_deg = 3
    MetaMod1.pce_q_norm = 1
    
    ExpDesign1 = ExpDesigns(Inputs)
    ExpDesign1.n_init_samples = 30
    ExpDesign1.sampling_method = 'random'
    
    Engine_1 = Engine(MetaMod1, Model1, ExpDesign1)
    Engine_1.train_normal()
    
    Model2 = PyLinkForwardModel()
    Model2.link_type = 'Function'
    Model2.py_file = 'model2'
    Model2.name = 'model2'
    Model2.Output.names = ['A']
    Model2.func_args = {'x_values': x_values}
    Model2.store = False
    Model2.observations = Model.observations
    
    MetaMod2 = MetaModel(Inputs)
    MetaMod2.meta_model_type = 'aPCE'
    MetaMod2.pce_reg_method = 'FastARD'
    MetaMod2.pce_deg = 3
    MetaMod2.pce_q_norm = 1
    
    ExpDesign2 = ExpDesigns(Inputs)
    ExpDesign2.n_init_samples = 30
    ExpDesign2.sampling_method = 'random'
    
    Engine_2 = Engine(MetaMod2, Model2, ExpDesign2)
    Engine_2.train_normal()
    
    meta_models = {
        "linear": Engine_,
        #"square": Engine_1,
        "degthree": Engine_2
        }

    BayesOpts = BayesModelComparison()
    
    opts_bootstrap = {
        "bootstrap": True,
        "n_samples": 10, #0,
        "Discrepancy": DiscrepancyOpts,
        "emulator": True,
        "plot_post_pred": False,
        "mcmc_params": {
            'n_steps': 100, #e4,
            'n_walkers': 10,
            'moves': emcee.moves.KDEMove(),
            'multiprocessing': False,
            'verbose': False
            }
        }

    output_dict = BayesOpts.model_comparison_all(
        meta_models,
        opts_bootstrap
        )